
//		FunctionHelp.cpp - plugin to add status-bar help for functions


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define PLUGIN_NAME					L"FunctionHelp"

#define VER_MAJOR					0
#define VER_MINOR					93
#define VER_BUILD					0
#define VER_PATCH					3

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif

#define MAX_ARG_LEN					8192
#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			32768


#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )


#define MAX_FUNCTION_NAME_LEN		48
#define MAX_FUNCTION_NESTING		64

#define TAB							0x09
#define LF							0x0a
#define NOT_SIGN					0xac
#define PARAGRAPH_MARK				0xb6



HINSTANCE MyDLLHandle				= NULL;



struct HelpNode {
	HelpNode *Before;
	HelpNode *After;
	LPSTR Syntax;
	LPSTR MoreInfo;
	wchar_t FuncName;
};

struct HelpAry {
	wchar_t FuncName[MAX_FUNCTION_NAME_LEN];
	LPSTR Syntax;
	LPSTR MoreInfo;
};

struct _FHMemInfo {
	unsigned int Nodes;
	size_t NodeStructures;
	size_t FunctionNames;
	size_t Syntax;
	size_t MoreInfo;
	size_t Total;
};

struct BracketEntry {
	wchar_t *Position;
	BOOL Function;
};




const wchar_t PluginFeatures[] = L"funchelp,*key";



const wchar_t FuncHelpHelpText[]	= L"Reload the FunctionHelp data file, or display info.\n\
\n\
FUNCHELP /I /L:filename /Q\r\n\
\n\
\t/I\tshow detailed memory usage  \n\
\t/L:\treload the data file\n\
\t/Q\tquietly\n\
\n\
If you load help text with /L, the text from the file will \n\
replace all currently loaded help text.\n";



const wchar_t EnvVarName[]			= L"FunctionHelp";
const wchar_t DefaultFileName[]		= L"FunctionHelp.txt";
const wchar_t StatusBarClass[]		= L"XTPStatusBar";
const wchar_t HelpTextPrefix[]		= L"\u200b";			//	zero width space to prevent USAGE from swallowing leading @ sign



HelpNode *HelpRoot					= NULL;
unsigned int HelpTreeNodes			= 0;
LPTSTR OldStatusBarText				= NULL;
BOOL GotStatusBarText				= FALSE;




int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



#include "ParseArgs.cpp"
#include "mmfiles.cpp"
#include "PluralStr.cpp"



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ch == ':' ) || ( ch == '=' );
}



int AppendFilename( LPTSTR Path, LPCTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



BOOL IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



int DeleteHelpTree( HelpNode *RootNode, unsigned int *Deleted )
{
	if ( !RootNode )
		return 0;


	if ( RootNode->Syntax ) {
		free( RootNode->Syntax );
		RootNode->Syntax = NULL;
	}

	if ( RootNode->MoreInfo ) {
		free( RootNode->MoreInfo );
		RootNode->MoreInfo = NULL;
	}

	if ( RootNode->Before )
		DeleteHelpTree( RootNode->Before, Deleted );

	if ( RootNode->After )
		DeleteHelpTree( RootNode->After, Deleted );

	free( RootNode );

	if ( Deleted )
		*Deleted += 1;

	return 0;
}



HelpNode *FindNode( HelpNode *StartNode, LPTSTR FuncName )
{
	if ( !StartNode )
		return NULL;


	int cmp		= wcscmp( FuncName, &( StartNode->FuncName ) );
	//	FmtMsgBox( 0, L"FindNode()", L"Compare '%s' to '%s':  %d\n", FuncName, &( StartNode->FuncName ), cmp );


	if ( cmp == 0 )
		return StartNode;
	else if ( cmp < 0 )
		return FindNode( StartNode->Before, FuncName );
	else
		return FindNode( StartNode->After, FuncName );
}



BOOL FuncNameChar( wchar_t ch )
{
	if ( __iswcsym( ch ) || ( ch == '$' ) )
		return TRUE;
	else
		return FALSE;
}



int ParseLineFromFile( LPTSTR Line, LPTSTR FuncName, LPTSTR Syntax, LPTSTR MoreInfo )
{
	if ( !Line || !FuncName )
		return -1;


	wchar_t *ch			= Line;
	int Brackets		= 0;
	int j				= 0;



	FuncName[0]			= '\0';

	while ( iswspace( *ch ) )
		ch++;

	if ( *ch != '@' )
		return 0;

	ch++;
	if ( !iswalpha( *ch ) )
		return 0;

	while ( FuncNameChar( *ch ) ) {
		if ( j >= ( MAX_FUNCTION_NAME_LEN - 1 ) ) {
			FuncName[0] = '\0';
			return 0;
		}

		FuncName[j++] = towupper( *ch );
		ch++;
	}

	if ( *ch != '[' ) {
		FuncName[0] = '\0';
		return 0;
	}

	FuncName[j] = '\0';

	if ( !Syntax && !MoreInfo )
		return j;
	

	if ( Syntax )
		Syntax[0]	= '\0';

	if ( MoreInfo )
		MoreInfo[0]	= '\0';


	do {
		if ( *ch == '[' )
			Brackets++;
		else if ( *ch == ']' )
			Brackets--;

		ch++;
	}
	while ( *ch && Brackets );

	if ( Syntax ) {
		wcsncpy_s( Syntax, SHORT_BUF_LEN, Line, ch - Line );
		for ( int i = 1; i <= j; i++ )
			Syntax[i] = towupper( Syntax[i] );
	}

	if ( MoreInfo && *ch ) {
		while ( iswspace( *ch ) )
			ch++;

		if ( *ch == '-' ) {
			while ( *ch == '-' )
				ch++;
		}
		else if ( ( *ch == ':' ) || ( ( *ch >= 0x2010 ) && ( *ch <= 0x2015 ) ) )
			ch++;

		while ( iswspace( *ch ) )
			ch++;

		if ( *ch )
			wcscpy_s( MoreInfo, MAX_ARG_LEN, ch );
	}

	return j;
}



int _cdecl HelpAryCmp$( const void *arg1, const void *arg2 )
{
	return wcscmp( ( (HelpAry*) arg1 )->FuncName, ( (HelpAry*) arg2 )->FuncName );
}



HelpNode *ArrayToTree( HelpAry *Array, int First, int Last, int *NodesAdded )
{
	HelpNode *NewNode	= NULL;
	int Middle			= First + ( Last - First + 1 ) / 2;


	NewNode = (HelpNode*) malloc( sizeof( HelpNode ) + ( wcslen( Array[Middle].FuncName ) * sizeof( wchar_t ) ) );
	if ( !NewNode )
		return NULL;

	wcscpy_s( &( NewNode->FuncName ), wcslen( Array[Middle].FuncName ) + 1, Array[Middle].FuncName );

	NewNode->Syntax		= Array[Middle].Syntax;
	NewNode->MoreInfo	= Array[Middle].MoreInfo;
	NewNode->Before		= NULL;
	NewNode->After		= NULL;

	if ( Middle > First ) {
		NewNode->Before = ArrayToTree( Array, First, Middle - 1, NodesAdded );

		if ( NewNode->Before == NULL ) {
			free( NewNode );
			return NULL;
		}
	}

	if ( Middle < Last ) {
		NewNode->After = ArrayToTree( Array, Middle + 1, Last, NodesAdded );

		if ( NewNode->After == NULL ) {
			free( NewNode );
			return NULL;
		}
	}

	if ( NodesAdded )
		*NodesAdded += 1;

	return NewNode;
}



HelpNode *LoadFileToTree( LPTSTR Filename, BOOL Quiet, unsigned int *NodesCount )
{
	wchar_t LineBuf[MAX_ARG_LEN]				= L"";
	wchar_t FuncName[MAX_FUNCTION_NAME_LEN]		= L"";
	wchar_t Syntax[SHORT_BUF_LEN]				= L"";
	char SyntaxUtf8[2 * SHORT_BUF_LEN]			= "";
	wchar_t MoreInfo[MAX_ARG_LEN]				= L"";
	char MoreInfoUtf8[2 * MAX_ARG_LEN]			= "";

	_MMFile MMFile;

	int ArraySize								= 0;
	int Count									= 0;
	int Found									= -1;
	
	HelpAry *HelpArray							= NULL;

	HelpNode *NewTree							= NULL;
	int Nodes									= 0;


	if ( OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES ) )
		return NULL;

	while ( MMGetLine( &MMFile, LineBuf, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) == 0 ) {
		if ( ParseLineFromFile( LineBuf, FuncName, NULL, NULL ) > 0 )
			ArraySize++;
	}

	if ( !ArraySize ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No functions found in file \"%s\".\n", Filename );
		goto Error;
	}


	HelpArray = (HelpAry*) malloc( ArraySize * sizeof( HelpAry ) );
	if ( !HelpArray ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Allocating help array\n" );
		goto Error;
	}


	MMFile.Index = MMFile.BOMSize;

	while ( MMGetLine( &MMFile, LineBuf, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) == 0 ) {

		if ( ParseLineFromFile( LineBuf, FuncName, Syntax, MoreInfo ) ) {

			Found = -1;

			SyntaxUtf8[0]	= '\0';
			MoreInfoUtf8[0]	= '\0';

			if ( Syntax[0] )
				WideCharToMultiByte( CP_UTF8, 0, Syntax, -1, SyntaxUtf8, 2 * SHORT_BUF_LEN, NULL, NULL );

			if ( MoreInfo[0] )
				WideCharToMultiByte( CP_UTF8, 0, MoreInfo, -1, MoreInfoUtf8, 2 * MAX_ARG_LEN, NULL, NULL );

			for ( int i = 0; i < Count; i++ )
				if ( !wcscmp( FuncName, HelpArray[i].FuncName ) )
					Found = i;

			if ( Found == -1 ) {
				wcscpy_s( HelpArray[Count].FuncName, MAX_FUNCTION_NAME_LEN, FuncName );
				HelpArray[Count].Syntax = (LPSTR) malloc( strlen( SyntaxUtf8 ) + 1 );
				if ( HelpArray[Count].Syntax == NULL )
					goto MallocStringsError;

				strcpy_s( HelpArray[Count].Syntax, strlen( SyntaxUtf8 ) + 1, SyntaxUtf8 );

				HelpArray[Count].MoreInfo = NULL;
				if ( MoreInfo[0] ) {
					HelpArray[Count].MoreInfo = (LPSTR) malloc( strlen( MoreInfoUtf8 ) + 1 );
					if ( HelpArray[Count].MoreInfo == NULL )
						goto MallocStringsError;

					strcpy_s( HelpArray[Count].MoreInfo, strlen( MoreInfoUtf8 ) + 1, MoreInfoUtf8 );
				}

				Count++;
			}
			else {
				free( HelpArray[Found].Syntax );
				free( HelpArray[Found].MoreInfo );

				HelpArray[Found].Syntax = (LPSTR) malloc( strlen( SyntaxUtf8 ) + 1 );
				if ( HelpArray[Found].Syntax == NULL )
					goto MallocStringsError;

				strcpy_s( HelpArray[Found].Syntax, strlen( SyntaxUtf8 ) + 1, SyntaxUtf8 );

				HelpArray[Found].MoreInfo = NULL;
				if ( MoreInfo[0] ) {
					HelpArray[Found].MoreInfo = (LPSTR) malloc( strlen( MoreInfoUtf8 ) + 1 );
					if ( HelpArray[Found].MoreInfo == NULL )
						goto MallocStringsError;

					strcpy_s( HelpArray[Found].MoreInfo, strlen( MoreInfoUtf8 ) + 1, MoreInfoUtf8 );
				}

			}
		}
	}

	CloseMMFile( &MMFile );


	if ( Count > 1 )
		qsort( HelpArray, Count, sizeof( HelpAry ), HelpAryCmp$ );


	Nodes = 0;

	NewTree = ArrayToTree( HelpArray, 0, Count - 1, &Nodes );
	if ( !NewTree ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error converting array to tree\n" );
		goto Error;
	}

	if ( !Quiet )
		Printf( L"Loaded %s from \"%s\".\r\n", PluralN( Nodes, L"function" ), Filename );

	if ( NodesCount )
		*NodesCount = Nodes;

	free( HelpArray );
	return NewTree;


Error:

	CloseMMFile( &MMFile );

	if ( HelpArray && Count ) {
		for ( int i = 0; i < Count; i++ ) {
			free( HelpArray[i].Syntax );
			free( HelpArray[i].MoreInfo );
		}
	}

	free( HelpArray );
	return NULL;


MallocStringsError:

	CloseMMFile( &MMFile );

	if ( HelpArray ) {
		for ( int i = 0; i <= Count; i++ ) {
			free( HelpArray[i].Syntax );
			free( HelpArray[i].MoreInfo );
		}

		free( HelpArray );
	}

	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Error allocating memory for help strings\r\n" );
	return NULL;
}



int _FH_GetMemInfo( HelpNode *Node, _FHMemInfo *Info )
{
	if ( Node ) {
		Info->Nodes += 1;
		Info->NodeStructures += sizeof( HelpNode );

		Info->FunctionNames += ( wcslen( &( Node->FuncName ) ) * sizeof( wchar_t ) );

		if ( Node->Syntax )
			Info->Syntax += ( strlen( Node->Syntax ) + 1 );

		if ( Node->MoreInfo )
			Info->MoreInfo += ( strlen( Node->MoreInfo ) + 1 );
	}

	if ( Node->Before )
		_FH_GetMemInfo( Node->Before, Info );

	if ( Node->After )
		_FH_GetMemInfo( Node->After, Info );

	return 0;
}



int GetMemoryInfo( HelpNode *HelpTree, _FHMemInfo *Info )
{
	if ( !Info )
		return -1;


	memset( Info, 0x00, sizeof( _FHMemInfo ) );

	if ( HelpTree ) {
		_FH_GetMemInfo( HelpTree, Info );
		Info->Total = Info->NodeStructures + Info->FunctionNames + Info->Syntax + Info->MoreInfo;
	}

	return Info->Nodes;
}



int ShowHelpText( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



int funchelp( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t HelpFile[MAX_FILENAME_LEN]		= L"";

	BOOL ShowMemInfo			= FALSE;
	BOOL Quiet					= FALSE;
	BOOL Unload					= FALSE;
	BOOL BasicInfo				= TRUE;

	HelpNode *NewTree			= NULL;
	unsigned int NewTreeNodes	= 0;
	unsigned int NodesDeleted	= 0;

	_FHMemInfo Info;

	int rv						= 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcscmp( arg[i], L"/?" ) )
			return ShowHelpText( FuncHelpHelpText, FALSE );
		else if ( !wcscmp( arg[i], L"/??" ) )
			return ShowHelpText( FuncHelpHelpText, TRUE );
		else if ( !_wcsicmp( arg[i], L"/I" ) )
			ShowMemInfo = TRUE, BasicInfo = FALSE;
		else if ( !_wcsicmp( arg[i], L"/L" ) ) {

			if ( i == ( numargs - 1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing filename: \"/L\"\r\n" );
				return 1;
			}
			else if ( arg[i + 1][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing filename: \"/L\"\r\n" );
				return 1;
			}

			i++;
			HelpFile[0] = '\0';
			QueryTrueName( arg[i], HelpFile );

			if ( HelpFile[0] == '\0' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( QueryIsDirectory( HelpFile ) )
				AppendFilename( HelpFile, DefaultFileName, MAX_FILENAME_LEN );

			if ( !IsAFile( HelpFile ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", HelpFile );
				return 2;
			}

			BasicInfo = FALSE;
		}
		else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			HelpFile[0] = '\0';
			QueryTrueName( arg[i] + j, HelpFile );

			if ( HelpFile[0] == '\0' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( QueryIsDirectory( HelpFile ) )
				AppendFilename( HelpFile, DefaultFileName, MAX_FILENAME_LEN );

			if ( !IsAFile( HelpFile ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", HelpFile );
				return 2;
			}

			BasicInfo = FALSE;
		}
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Quiet = TRUE;
		else if ( !_wcsicmp( arg[i], L"/Z" ) )
			Unload = TRUE, BasicInfo = FALSE;
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	if ( Unload ) {
		if ( HelpRoot ) {
			NodesDeleted	= 0;
			DeleteHelpTree( HelpRoot, &NodesDeleted );
			HelpRoot		= NULL;
			HelpTreeNodes	= 0;

			if ( !Quiet )
				Printf( L"\u2022  %s deleted.\r\n", PluralN( NodesDeleted, L"function" ) );
		}
		else {
			if ( !Quiet )
				Printf( L"Nothing to unload.\r\n" );
		}
	}


	if ( HelpFile[0] ) {
		NewTree = LoadFileToTree( HelpFile, Quiet, &NewTreeNodes );

		if ( NewTree ) {
			if ( HelpRoot ) {
				NodesDeleted	= 0;
				DeleteHelpTree( HelpRoot, &NodesDeleted );
				HelpRoot		= NULL;
				HelpTreeNodes	= 0;
			}

			HelpRoot = NewTree;
			HelpTreeNodes = NewTreeNodes;
		}
	}


	if ( ShowMemInfo ) {
		GetMemoryInfo( HelpRoot, &Info );

		Printf( L"\r\nFound %s:\r\n\r\n", PluralN( Info.Nodes, L"function" ) );
		Printf( L"Node structures:  %s\r\n", PluralNW( Info.NodeStructures, 9, L"byte" ) );
		Printf( L"Function names:   %s\r\n", PluralNW( Info.FunctionNames, 9, L"byte" ) );
		Printf( L"Function syntax:  %s\r\n", PluralNW( Info.Syntax, 9, L"byte" ) );
		Printf( L"Additional info:  %s\r\n", PluralNW( Info.MoreInfo, 9, L"byte" ) );
		Printf( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\r\n" );
		Printf( L"TOTAL:            %s\r\n", PluralNW( Info.Total, 9, L"byte" ) );
	}
	else if ( BasicInfo )
		Printf( L"%s in memory.\r\n", PluralN( HelpTreeNodes, L"function" ) );

	return rv;
}



int AutoLoadFile()
{
	wchar_t HelpFile[MAX_FILENAME_LEN]		= L"";
	wchar_t VarBuf[MAX_ARG_LEN]				= L"";
	wchar_t *ch								= NULL;


	//	First try in the program directory :

	wcscpy_s( HelpFile, MAX_FILENAME_LEN, L"%@path[%_cmdspec]" );
	ExpandVariables( HelpFile, 1 );
	AppendFilename( HelpFile, DefaultFileName, MAX_FILENAME_LEN );

	if ( IsAFile( HelpFile ) ) {
		HelpRoot = LoadFileToTree( HelpFile, TRUE, &HelpTreeNodes );

		if ( HelpRoot )
			return 0;
	}


	//	Then look in the DLL's directory :

	HelpFile[0]		= '\0';
	GetModuleFileName( MyDLLHandle, HelpFile, MAX_FILENAME_LEN );

	if ( HelpFile[0] ) {
		ch = wcsrchr( HelpFile, '\\' );
		if ( ch ) {
			ch[1] = '\0';
			wcscat_s( HelpFile, MAX_FILENAME_LEN, DefaultFileName );

			if ( IsAFile( HelpFile ) ) {
				HelpRoot = LoadFileToTree( HelpFile, TRUE, &HelpTreeNodes );

				if ( HelpRoot )
					return 0;
			}
		}
	}


	//	Finally, try the environment variable :

	HelpFile[0] = '\0';
	GetEnvironmentVariable( EnvVarName, VarBuf, MAX_ARG_LEN );

	if ( VarBuf[0] ) {
		QueryTrueName( VarBuf, HelpFile );

		if ( QueryIsDirectory( HelpFile ) )
			AppendFilename( HelpFile, DefaultFileName, MAX_FILENAME_LEN );

		if ( HelpFile[0] ) {

			if ( IsAFile( HelpFile ) ) {
				HelpRoot = LoadFileToTree( HelpFile, TRUE, &HelpTreeNodes );

				if ( HelpRoot )
					return 0;
			}
		}
	}

	return 0;
}



int ReplaceChar( LPTSTR String, wchar_t OldChar, wchar_t NewChar )
{
	if ( !String || !OldChar || !NewChar || ( OldChar == NewChar ) )
		return -1;


	int Count		= 0;


	for ( wchar_t *ch = String; *ch; ch++ )
		if ( *ch == OldChar )
			*ch = NewChar, Count++;

	return Count;
}



int CombineHelpText( HelpNode *Node, LPTSTR outBuffer )
{
	if ( !Node || !outBuffer )
		return -1;


	wchar_t Syntax[SHORT_BUF_LEN]		= L"";
	wchar_t MoreInfo[MAX_ARG_LEN]		= L"";


	outBuffer[0] = '\0';

	if ( Node->Syntax )
		MultiByteToWideChar( CP_UTF8, 0, Node->Syntax, -1, Syntax, SHORT_BUF_LEN );

	if ( Node->MoreInfo )
		MultiByteToWideChar( CP_UTF8, 0, Node->MoreInfo, -1, MoreInfo, MAX_ARG_LEN );

	if ( Syntax[0] ) {
		wcscpy_s( outBuffer, MAX_ARG_LEN, HelpTextPrefix );
		wcscat_s( outBuffer, MAX_ARG_LEN, Syntax );

		if ( MoreInfo[0] ) {
			ReplaceChar( MoreInfo, PARAGRAPH_MARK, LF );
			ReplaceChar( MoreInfo, NOT_SIGN, TAB );
			wcscat_s( outBuffer, MAX_ARG_LEN, L"\r\n\n" );
			wcscat_s( outBuffer, MAX_ARG_LEN, MoreInfo );
		}
	}

	return 0;
}



int FreeOldStatusBarText()
{
	free( OldStatusBarText );
	OldStatusBarText	= NULL;
	return 0;
}



BOOL CALLBACK _FindTCStatusBar( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return FALSE;


	HWND *FoundWindow					= (HWND*) lParam;
	wchar_t ClassName[SHORT_BUF_LEN]	= L"";


	RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN );
	if ( _wcsicmp( ClassName, StatusBarClass ) )
		return TRUE;

	*FoundWindow = hWnd;
	return FALSE;
}



int GetStatusBarText()
{
	if ( IsWindowVisible( GetConsoleWindow() ) )
		return 0;


	wchar_t shortbuf[SHORT_BUF_LEN]			= L"";
	HWND TakeCmdWindow						= NULL;
	HWND StatusBar							= NULL;


	TakeCommandIPC( L"HWND", shortbuf );
	if ( shortbuf[0] == '\0' )
		return 0;


	_set_errno( 0 );
	TakeCmdWindow = (HWND) wcstoul( shortbuf, NULL, 10 );
	if ( errno || !TakeCmdWindow )
		return 0;


	EnumChildWindows( TakeCmdWindow, _FindTCStatusBar, (LPARAM) &StatusBar );
	if ( !StatusBar )
		return 0;


	shortbuf[0] = '\0';
	SendMessage( StatusBar, WM_GETTEXT, SHORT_BUF_LEN, (LPARAM) shortbuf );


	FreeOldStatusBarText();

	if ( shortbuf[0] ) {
		OldStatusBarText = (LPTSTR) malloc( ( wcslen( shortbuf ) + 1 ) * sizeof( wchar_t ) );
		if ( OldStatusBarText )
			wcscpy_s( OldStatusBarText, wcslen( shortbuf ) + 1, shortbuf );
	}

	return 0;
}



int FunctionBeforeCursor( LPTSTR CommandLine, size_t Cursor, LPTSTR FuncName )
{
	if ( !CommandLine || !FuncName )
		return -1;


	wchar_t *Pct		= NULL;
	wchar_t *Start		= NULL;
	size_t len			= 0;



	FuncName[0] = '\0';

	for ( wchar_t *ch = CommandLine; ch < ( CommandLine + Cursor ); ch++ )
		if ( *ch == '%' )
			Pct = ch;

	if ( !Pct )
		return 0;

	if ( Pct[1] != '@' )
		return 0;

	Start = Pct + 2;
	if ( !iswalpha( *Start ) )
		return 0;

	len = ( CommandLine + Cursor ) - Start;
	if ( len >= MAX_FUNCTION_NAME_LEN )
		return 0;

	for ( size_t i = 0; i < len; i++ ) {
		if ( !FuncNameChar( Start[i] ) ) {
			FuncName[0] = '\0';
			return 0;
		}

		FuncName[i] = towupper( Start[i] );
	}

	FuncName[len] = '\0';
	return (int) len;
}



int HandleOpenBracket( LPKEYINFO ki )
{
	if ( !HelpRoot )
		return 0;

	if ( IsWindowVisible( GetConsoleWindow() ) )
		return 0;


	wchar_t FuncName[MAX_FUNCTION_NAME_LEN]		= L"";
	wchar_t HelpText[MAX_ARG_LEN]				= L"";

	HelpNode *Node				= NULL;


	FunctionBeforeCursor( ki->pszLine, ki->pszCurrent - ki->pszLine, FuncName );

	if ( FuncName[0] == '\0' )
		return 0;


	Node = FindNode( HelpRoot, FuncName );
	if ( !Node )
		return 0;

	CombineHelpText( Node, HelpText );

	if ( !GotStatusBarText ) {
		GetStatusBarText();
		GotStatusBarText = TRUE;
	}

	TakeCommandIPC( L"USAGE", HelpText );
	return 0;
};



int HandleCloseBracket( LPKEYINFO ki )
{
	if ( !HelpRoot )
		return 0;

	if ( IsWindowVisible( GetConsoleWindow() ) )
		return 0;

	
	BracketEntry BracketStack[MAX_FUNCTION_NESTING];
	int Depth			= 0;
	int Overflow		= 0;

	wchar_t FuncName[MAX_FUNCTION_NAME_LEN]		= L"";
	wchar_t HelpText[MAX_ARG_LEN]				= L"";
	
	HelpNode *Node				= NULL;
	BOOL UseOldStatus			= FALSE;


	for ( wchar_t *ch = ki->pszLine; ch < ki->pszCurrent; ch++ ) {

		if ( *ch == '[' ) {

			if ( Depth < MAX_FUNCTION_NESTING ) {

				FuncName[0] = '\0';
				FunctionBeforeCursor( ki->pszLine, ch - ki->pszLine, FuncName );

				BracketStack[Depth].Position	= ch;
				BracketStack[Depth].Function	= ( FuncName[0] != '\0' );
				Depth++;
			}
			else
				Overflow++;

		}
		else if ( *ch == ']' ) {

			if ( Overflow )
				Overflow--;
			else if ( Depth )
				Depth--;
		}
	}


	if ( Overflow )
		return 0;

	UseOldStatus = TRUE;

	if ( !Depth )
		goto Abort;

	//	Pop the last item from the stack.  (The user just typed a close-bracket, closing it.)

	Depth--;
	if ( !Depth )
		goto Abort;

	//	Pop any non-function brackets from the stack :

	while ( BracketStack[Depth - 1].Function == FALSE ) {
		Depth--;
		if ( !Depth )
			goto Abort;
	}

	FuncName[0] = '\0';

	if ( Depth ) {
		if ( BracketStack[Depth - 1].Function ) {
			FunctionBeforeCursor( ki->pszLine, BracketStack[Depth - 1].Position - ki->pszLine, FuncName );

			if ( FuncName[0] ) {
				Node = FindNode( HelpRoot, FuncName );

				if ( Node )
					CombineHelpText( Node, HelpText );
			}
		}
	}

Abort:

	if ( HelpText[0] )
		TakeCommandIPC( L"USAGE", HelpText );
	else if ( UseOldStatus )
		TakeCommandIPC( L"USAGE", OldStatusBarText );
	else
		TakeCommandIPC( L"USAGE", NULL );

	return 0;
}



DLLExports int WINAPI key( LPKEYINFO ki )
{

	if ( ( ki->nKey == 0 ) && ( ki->pszLine == NULL ) ) {
		FreeOldStatusBarText();
		GotStatusBarText = FALSE;
	}
	else if ( ki->nKey == '[' )
		HandleOpenBracket( ki );
	else if ( ki->nKey == ']' )
		HandleCloseBracket( ki );
	
	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/plugin.html";
	piInfo.pszDescription	= (LPTSTR) L"Provides status-bar help for functions.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]";
	ExpandVariables( shortbuf, 1 );


	if ( wcscmp( shortbuf, L"00" ) )
		return 1;


	shortbuf[0] = '\0';
	GetEnvironmentVariable( L"NoLoadMsg", shortbuf, SHORT_BUF_LEN );


	AutoLoadFile();


	if ( shortbuf[0] == '\0' )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );


	if ( HelpRoot ) {
		if ( shortbuf[0] == '\0' )
			Printf( L"\u2022  %s loaded into memory.\r\n", PluralN( HelpTreeNodes, L"function" ) );
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s not found!\r\n", DefaultFileName );
		Qprintf( ERRHANDLE, L"Function help is not (yet) available.  Use FUNCHELP /L to load it.\r\n" );
	}

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";
	unsigned int NodesDeleted			= 0;


	GetEnvironmentVariable( L"NoLoadMsg", shortbuf, SHORT_BUF_LEN );


	FreeOldStatusBarText();

	if ( HelpRoot ) {
		NodesDeleted	= 0;
		DeleteHelpTree( HelpRoot, &NodesDeleted );
		HelpRoot		= 0;
		HelpTreeNodes	= 0;

		if ( shortbuf[0] == '\0' )
			Printf( L"\u2022  %s deleted.\r\n", PluralN( NodesDeleted, L"function" ) );
	}

	return 0;
}



const int NumHelpTexts							= 1;
const wchar_t HelpFeature[NumHelpTexts][11]		= { L"funchelp" };
const wchar_t *HelpText[NumHelpTexts]			= { FuncHelpHelpText };



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}



/*

		0.92.0.1	2021-09-12

		0.92.1.0	2021-12-21
					Fix an issue calling TakeCommandIPC() from 32-bit TCC.

		0.92.2.0	2021-12-21
					Why was I ever dynamically linking to that function in the
					first place?  Stop the madness!

		0.93.0.0	2023-06-16
					Removed all calls to QueryIsFile() to work around an issue
					in TCC v30.00.20.

*/
